#!/usr/bin/env python3
"""
yolo_sliced_geotiff_to_shp.py

Run Ultralytics YOLO on a very large georeferenced TIFF by slicing it into overlapping tiles,
mapping detections back to geospatial coordinates, performing a global class-aware NMS,
and exporting results to a Shapefile (.shp).

Usage:
    python yolo_sliced_geotiff_to_shp.py --tif path/to/large.tif --weights yolov8n.pt --out out_detections.shp
"""

import argparse
import math
import os
from pathlib import Path
import numpy as np
import cv2
from ultralytics import YOLO
import torch
from torchvision.ops import batched_nms
import rasterio
from rasterio.windows import Window
from rasterio.transform import Affine
from shapely.geometry import box as shapely_box
import geopandas as gpd
from tqdm import tqdm
import rasterio
from rasterio.windows import Window
from rasterio.transform import Affine, xy  # <- add xy here


# ----------------------------
# helpers
# ----------------------------
def make_tiles(img_w, img_h, tile_size, overlap):
    """Return list of (x1,y1,x2,y2) pixel coords covering image with overlap (pixels)."""
    step = tile_size - overlap
    nx = max(1, math.ceil((img_w - overlap) / step))
    ny = max(1, math.ceil((img_h - overlap) / step))

    tiles = []
    for iy in range(ny):
        for ix in range(nx):
            x1 = ix * step
            y1 = iy * step
            x2 = x1 + tile_size
            y2 = y1 + tile_size

            if x2 > img_w:
                x2 = img_w
                x1 = max(0, x2 - tile_size)
            if y2 > img_h:
                y2 = img_h
                y1 = max(0, y2 - tile_size)

            tiles.append((int(x1), int(y1), int(x2), int(y2)))
    # remove duplicates
    tiles = list(dict.fromkeys(tiles))
    return tiles

def read_tile_rgb(dataset, x1, y1, x2, y2, out_size=None):
    """
    Read window [x1,x2) x [y1,y2) from rasterio dataset and return HWC RGB uint8 array.
    out_size: (height, width) to rescale the read tile to (optional).
    Assumes raster has at least 3 bands (RGB). If >3 bands, uses first 3.
    """
    w = x2 - x1
    h = y2 - y1
    window = Window(x1, y1, w, h)
    # Read bands (1-indexed). Read minimum of 3 bands or all if fewer
    count = dataset.count
    bands_to_read = min(3, count)
    arr = dataset.read(range(1, bands_to_read + 1), window=window)  # shape (bands, h, w)

    # If bands < 3, replicate or pad
    if arr.shape[0] == 1:
        arr = np.repeat(arr, 3, axis=0)
    elif arr.shape[0] == 2:
        arr = np.vstack([arr, arr[0:1, :, :]])

    # convert to HWC
    tile = np.transpose(arr, (1, 2, 0))  # H, W, C

    # Raster values may be not in 0-255. Attempt rescale if dtype != uint8
    if tile.dtype != np.uint8:
        # scale using min/max per-tile (safe heuristic). If you want to use dataset-specific scaling, change here.
        tmin, tmax = tile.min(), tile.max()
        if tmax > tmin:
            tile = ((tile - tmin) / (tmax - tmin) * 255.0).astype(np.uint8)
        else:
            tile = np.clip(tile, 0, 255).astype(np.uint8)
    else:
        tile = tile.copy()

    # If out_size specified, resize tile to that size (width,height)
    if out_size is not None:
        tile = cv2.resize(tile, (out_size[1], out_size[0]), interpolation=cv2.INTER_LINEAR)

    # Ensure channel order is RGB (rasterio gives bands in dataset order)
    return tile

def pixel_to_geo(transform: Affine, x, y):
    """Map pixel coordinates (x,y) to geographic coords using rasterio transform (upper-left origin)."""
    # rasterio.transform.xy expects row, col -> y, x (row = y pixel)
    # but we pass x,y as pixel coords so row=y, col=x
    # xy returns (x_geo, y_geo)
    from rasterio.transform import xy
    xgeo, ygeo = xy(transform, y,x, offset='center')
    return xgeo, ygeo

# ----------------------------
# main tiled inference -> shapefile
# ----------------------------
def run_geotiff_sliced_to_shp(
    tif_path,
    weights="yolov8n.pt",
    tile_size=512,
    overlap=200,
    conf_thresh=0.25,
    iou_thresh=0.45,
    device=None,
    out_shp="detections.shp",
    class_names=None,
    debug=False
):
    tif_path = str(tif_path)
    if not os.path.exists(tif_path):
        raise FileNotFoundError(tif_path)

    # open dataset
    with rasterio.open(tif_path) as ds:
        img_w = ds.width
        img_h = ds.height
        transform = ds.transform
        crs = ds.crs  # may be None

        tiles = make_tiles(img_w, img_h, tile_size, overlap)
        print(f"[info] image: {img_w}x{img_h}, tiles: {len(tiles)}, tile={tile_size}, overlap={overlap}")
        model = YOLO(weights)
        if device:
            model.model.to(device)

        all_boxes = []   # pixel-space boxes [x1,y1,x2,y2]
        all_scores = []
        all_classes = []

        for t_idx, (x1, y1, x2, y2) in enumerate(tqdm(tiles)):
            # Read tile into RGB np.uint8. Resize to model-friendly size if desired (we let model auto-resize)
            tile_rgb = read_tile_rgb(ds, x1, y1, x2, y2, out_size=None)  # HWC uint8
            if tile_rgb is None:
                continue

            # Ultralytics model.predict accepts numpy HWC (RGB)
            # we pass imgsz=min(tile_size,2048) and conf threshold
            results = model.predict(source=tile_rgb, conf=conf_thresh, imgsz=min(tile_size, 2048), device=device, verbose=False)
            r = results[0]

            if r.boxes is None or len(r.boxes) == 0:
                if debug:
                    print(f" tile {t_idx+1}/{len(tiles)}: 0 boxes")
                continue

            boxes = r.boxes.xyxy.cpu().numpy()  # coordinates in tile pixel space
            scores = r.boxes.conf.cpu().numpy()
            classes = r.boxes.cls.cpu().numpy().astype(int)

            # map tile coords -> full image pixel coords by adding offsets
            boxes[:, 0] += x1  # x1
            boxes[:, 1] += y1  # y1
            boxes[:, 2] += x1  # x2
            boxes[:, 3] += y1  # y2

            for b, s, c in zip(boxes, scores, classes):
                all_boxes.append(b.tolist())
                all_scores.append(float(s))
                all_classes.append(int(c))

            if debug:
                print(f" tile {t_idx+1}/{len(tiles)}: {len(boxes)} boxes")

    # no detections
    if len(all_boxes) == 0:
        print("[info] no detections found")
        return

    # convert to tensors and clamp within image bounds
    boxes = torch.tensor(all_boxes, dtype=torch.float32)
    scores = torch.tensor(all_scores, dtype=torch.float32)
    labels = torch.tensor(all_classes, dtype=torch.int64)

    # clamp boxes to image
    boxes[:, 0].clamp_(0, img_w - 1)
    boxes[:, 1].clamp_(0, img_h - 1)
    boxes[:, 2].clamp_(0, img_w - 1)
    boxes[:, 3].clamp_(0, img_h - 1)

    # Perform class-aware NMS
    keep = batched_nms(boxes, scores, labels, iou_thresh)
    boxes_kept = boxes[keep].cpu().numpy()
    scores_kept = scores[keep].cpu().numpy()
    labels_kept = labels[keep].cpu().numpy()

    print(f"[info] {len(all_boxes)} raw boxes -> {len(keep)} after global NMS")

    # Build GeoDataFrame with bbox polygons
    geoms = []
    attrs = []

    for (x1, y1, x2, y2), s, cls in zip(boxes_kept, scores_kept, labels_kept):

        # Convert BOX EDGES → PIXEL CENTERS
        print(y1)
        print(y2)
       
        x1g, y1g = xy(transform, [y1], [x1], offset='ul')
        x2g, y2g = xy(transform, [y2], [x2], offset='ul')

        print(y1g)
        print(y2g)

        xmin, xmax = sorted([x1g, x2g])
        ymin, ymax = sorted([y1g, y2g])
        geoms.append(shapely_box(xmin, ymin, xmax, ymax))

        attrs.append({
            "class_id": int(cls),
            "score": float(s)
        })

    gdf = gpd.GeoDataFrame(attrs, geometry=geoms, crs=crs)

    # If class names provided, add column
    if class_names is not None:
        # create name column mapping through IDs
        gdf["class_name"] = gdf["class_id"].apply(lambda x: class_names[x] if 0 <= x < len(class_names) else str(x))

    # Save to shapefile (driver determined by extension)
    out_shp = str(out_shp)
    out_dir = os.path.dirname(out_shp)
    if out_dir and not os.path.exists(out_dir):
        os.makedirs(out_dir, exist_ok=True)

    # geopandas will save as shapefile when .shp extension provided. Note: shapefile has 10-char limits for field names.
    print(f"[info] writing {len(gdf)} features to {out_shp} (CRS={crs})")
    gdf.to_file(out_shp, driver="ESRI Shapefile")
    print("[info] done.")


# ----------------------------
# CLI
# ----------------------------
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--tif", required=True, help="Path to large georeferenced TIFF")
    parser.add_argument("--weights", default="yolov8n.pt", help="YOLO weights")
    parser.add_argument("--tile_size", type=int, default=1024)
    parser.add_argument("--overlap", type=int, default=200)
    parser.add_argument("--conf_thresh", type=float, default=0.25)
    parser.add_argument("--iou_thresh", type=float, default=0.45)
    parser.add_argument("--device", default=None, help="'cpu' or 'cuda:0'")
    parser.add_argument("--out", default="detections.shp", help="Output shapefile (.shp)")
    parser.add_argument("--class_names", default=None, help="Optional path to newline-separated class names file")
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()

    class_names = None
    if args.class_names:
        with open(args.class_names, "r") as f:
            class_names = [ln.strip() for ln in f if ln.strip()]

    run_geotiff_sliced_to_shp(
        tif_path=args.tif,
        weights=args.weights,
        tile_size=args.tile_size,
        overlap=args.overlap,
        conf_thresh=args.conf_thresh,
        iou_thresh=args.iou_thresh,
        device=args.device,
        out_shp=args.out,
        class_names=class_names,
        debug=args.debug
    )




import requests
from bs4 import BeautifulSoup
import re
from datetime import datetime
from steren.network import Detector
import geopandas as gpd
from shapely.geometry import box
import os
from shapely.geometry import Point
import math



headers = {
    "User-Agent": "Mozilla/5.0 (compatible; research-scraper/1.0)"
}

def clean_text(s):
    if s is None:
        return None
    return (
        s.replace("\xa0", " ")
         .replace("°", "")
         .strip()
    )


def extract_all_metadata(soup):
    """
    Extracts all metadata fields from a HiRISE details page.
    Returns a dict: {label: value}
    """
    metadata = {}
    
    # Find all <strong> tags inside the main details container
    for strong in soup.find_all("strong"):
        label = clean_text(strong.get_text())
        
        # Skip empty labels
        if not label:
            continue
        
        # Find the value after the <strong>
        value = None
        for sib in strong.next_siblings:
            if isinstance(sib, str):
                value_candidate = sib.strip()
                if value_candidate:
                    value = value_candidate
                    break
            elif hasattr(sib, "get_text"):
                value_candidate = sib.get_text(strip=True)
                if value_candidate:
                    value = value_candidate
                    break
        
        if value:
            metadata[label] = clean_text(value)
    
    return metadata

def human_size(size_bytes):
    """Convert size in bytes (int) to human-readable format"""
    for unit in ['B','KB','MB','GB','TB']:
        if size_bytes < 1024:
            return f"{size_bytes:.1f}{unit}"
        size_bytes /= 1024
    return f"{size_bytes:.1f}PB"

class HRimage:
    
    def __init__(self, identifier : str = None, img_path : str = None):

        self.identifier = identifier 
        if identifier:
            self.mission_phase, orbit_number, latitude10 = identifier.split('_')
            self.orbit_number = int(orbit_number)
            self.latitude = int(latitude10)/10
            self.url = "https://uahirise.org/" + identifier

        if img_path:
            self.img_path = img_path


    def meta(self):
        """ Scrapes web for image metadata"""
        resp = requests.get(self.url,headers = headers, timeout = 30)
        resp.raise_for_status()

        soup = BeautifulSoup(resp.text,'lxml')
        title_tag = soup.select_one("span.observation-title-milo")
        self.title = title_tag.get_text(strip=True) if title_tag else None
        
        desc_raw = soup.select_one("div.caption-text-no-extras")
        self.description = clean_text(desc_raw.get_text(" ", strip=True) if desc_raw else None)
        
        raw_metadata = extract_all_metadata(soup)

        self.date = raw_metadata.get("Acquisition date", None)
        self.latitude = float(raw_metadata.get("Latitude (centered)", None))
        self.longitude = float(raw_metadata.get("Longitude (East)", None))
        self.altitude = raw_metadata.get("Spacecraft altitude", None)



        return {
            "observations_id" : self.identifier,
            "title" : self.title,
            "description": self.description,
            "acquisition date" : self.date,
            "latitude" : self.latitude,
            "longitude": self.longitude,
            "spacecraft altitude": self.altitude
            }

    def download(self, output_dir: str):
        """List PDS files with size and date, let user pick one"""
        orbit_range = f"{self.orbit_number - (self.orbit_number % 100):06d}_{self.orbit_number - (self.orbit_number % 100) + 99:06d}"
        url = f"https://hirise-pds.lpl.arizona.edu/PDS/EXTRAS/RDR/{self.mission_phase}/ORB_{orbit_range}/{self.identifier}/"
        print(f"Fetching: {url}\n")

        r = requests.get(url)
        if r.status_code != 200:
            print("Failed to access PDS directory.")
            return None

        soup = BeautifulSoup(r.text, "html.parser")

        # PDS directory listing is inside <pre>
        pre = soup.find("pre")
        if not pre:
            print("No directory listing found.")
            return None

        files = []
        for line in pre.text.splitlines():
            parts = line.split()
            if len(parts) < 2 or line.startswith("../"):
                continue
            filename = parts[0]
            date = parts[1] + " " + parts[2] if len(parts) >= 3 else ""
            size = None
            # last numeric part is usually the size
            for p in reversed(parts):
                if p.isdigit():
                    size = int(p)
                    break
            files.append((filename, date, human_size(size)))

        # print list
        for i, (fname, date, size) in enumerate(files, 1):
            print(f"{i}. {fname:40} {date:20} {size:>8}")

        choice = input(f"\nSelect a file (1-{len(files)}): ")
        if not choice.isdigit() or not (1 <= int(choice) <= len(files)):
            print("Invalid selection.")
            return None

        selected = files[int(choice)-1][0]
        full_url = url + selected
        print(f"\nSelected: {selected}\nURL: {full_url}")
        return full_url

    def detect(self, detector: Detector):
        """Runs a detector on the image and stores the results"""
        # add a condition here where self.img_path must have a value and exist? 
        if self.img_path is None:
            raise ValueError("Please provide an image path.")

        img_path = os.path.exists(self.img_path)
        if not os.path.exists(self.img_path):
            raise FileNotFoundError(f"Image path does not exist: {img_path}")


        boxes, scores, classes, crs, transform = detector.inference(self.img_path)
        self.boxes = boxes
        self.scores = scores
        self.classes = classes
        self.crs = crs
        self.transform = transform
    
    def export_detections(self, output_path):

        if not hasattr(self, "boxes") or self.boxes is None:
            raise RuntimeError("No detections found. Run detect() first.")

        # Check file extension
        if not output_path.lower().endswith(".gpkg"):
            raise ValueError("Output path must end with .gpkg")


        geometries = []
        diameters = []
        for xmin, ymin, xmax, ymax in self.boxes:
            # pixel → map coords
            x1, y1 = self.transform * (xmin, ymin)
            x2, y2 = self.transform * (xmax, ymax)

            # center of box (map coords)
            cx = (x1 + x2) / 2.0
            cy = (y1 + y2) / 2.0

            # average radius in map units
            radius = (abs(x2-x1)+abs(y2-y1))/4
            diameters.append(2*radius)

            # create circular polygon
            geometries.append(Point(cx, cy).buffer(radius))

        self.diameters = diameters



        gdf = gpd.GeoDataFrame(
            {
                "class": self.classes,
                "score": self.scores,
                "diameter": self.diameters,
            },
            geometry=geometries,
            crs=self.crs,
        )

        # Ensure output directory exists
        output_dir = os.path.dirname(output_path)
        if output_dir and not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)

        # Write GeoPackage
        gdf.to_file(
            output_path,
            layer="detections",
            driver="GPKG") 



    def search(self):
        hirise_catalog = "https://www.uahirise.org/catalog/index.php?page="
        page_num = 0
        while True:
            page_num += 1 
            url = hirise_catalog + str(page_num)
            r = requests.get(url)
            
            if r.status_code != 200:
                print("Failed to access HiRISE catalog.")
                return None
            
            html = r.text
            codes = sorted(set(re.findall(r'ESP_\d{6}_\d{4}', html)))
            for identifier in codes:
                hrimage = HRimage(identifier)
                hrimage.meta()
                print(f"{hrimage.identifier:20} {hrimage.title:60} {hrimage.date:>10}")
                







# src/gwel/cli.py
import os
import json
import yaml
import typer
from steren.data import HRimage



app = typer.Typer(add_completion = True, help="STEREN CLI tool", invoke_without_command=True, context_settings={"help_option_names": ["-h", "--help"]})

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    from importlib_metadata import version, PackageNotFoundError  # For Python <3.8

try:
    VERSION = version("steren")
except PackageNotFoundError:
    VERSION = "unknown"


ASCII_ART = r"""
   _____________________  _______   __
  / ___/_  __/ ____/ __ \/ ____/ | / /
  \__ \ / / / __/ / /_/ / __/ /  |/ / 
 ___/ // / / /___/ _, _/ /___/ /|  /  
/____//_/ /_____/_/ |_/_____/_/ |_/   
                                      """


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    version_flag: bool = typer.Option(
        False, "--version", "-v", help="Show STEREN CLI version"
    )):
    """
    Callback function that runs when 'steren' is called without a subcommand.
    """

    if version_flag:
        typer.secho(f"STEREN CLI version {VERSION}", fg=typer.colors.GREEN, bold=True)
        raise typer.Exit()

    if ctx.invoked_subcommand is None:
        print(ASCII_ART)
        typer.secho(f"STEREN CLI version {VERSION}", fg=typer.colors.CYAN, bold=True)
        typer.echo("Available commands:")

        # Use rich-like table formatting with Typer
        for name, cmd in ctx.command.commands.items():
            description = cmd.help or cmd.callback.__doc__ or ""
            if description:
                description = description.strip().split("\n")[0]
            typer.secho(f"  {name:<10}", fg=typer.colors.YELLOW, bold=True, nl=False)
            typer.echo(f" {description}")

@app.command()
def search(
    ctx: typer.Context, 
        type: str= typer.Argument(help="Type of planetary data [Supported types: HiRISE].")):
    """
    Search planetary data catalogues.
    """
    match type:
        case 'HiRISE':
            HRimage().search()


@app.command()
def metadata(
    ctx: typer.Context,
    type: str= typer.Argument(help="Type of planetary data [Supported types: HiRISE]."),
    identifier: str= typer.Argument(help ="Identifier of the planetary data")
    ):
    """
    Obtain planetary data metadata.
    """
    match type:
        case "HiRISE":
            hrimage = HRimage(identifier)
            metadata = hrimage.meta()
            print(yaml.dump(metadata, default_flow_style=False))
        case _:
            typer.secho(f"Data type {type} is not supported.")
         


@app.command()
def download(
    ctx: typer.Context,
    type: str= typer.Argument(help="Type of planetary data [Supported types: HiRISE]."),
    identifier: str= typer.Argument(help ="Identifier of the planetary data"),
    output_dir: str= typer.Option(".","-o","--output_dir",help="Directory path to download data into.")
        ):
    """
    Download planetary data.
    """
    match type:
        case "HiRISE":
            hrimage = HRimage(identifier)
            hrimage.download(".")
        case _:
            typer.secho(f"Data type {type} is not supported.")
       

@app.command()
def detect(
    ctx: typer.Context,
    data_type: str= typer.Argument(help="Type of planetary data [Supported types: HiRISE]."),
    detector_type: str= typer.Argument(help="Type of detector [Supported types: YOLOv8]."),
    model_weights: str=typer.Argument(help="Path to model weights."),
    img_path: str=typer.Argument(help="Path to image data."),
    output_file: str=typer.Argument(help="Output file path [Supported types: Geopackage (.gpkg), Shapefile (.shp), Tabular (.csv)]"),
    tile_size: int= typer.Option(512,'-ts','--tile_size', help="Tile size"),
    overlap: int= typer.Option(100,'-o','--overlap', help="Overlap size"),
    conf_thresh: float=typer.Option(0.25,'-c','--confidence',help="Minimum confidence threshold of detections."),
    iou_thresh: float=typer.Option(0.45, '-iou','--intersection_over_union', help="Threshold for the maximum intersection over union between detections.")):
    """
    Detect instances within planetary data with a pre-trained model.
    """
    match (data_type, detector_type):
        case ('HiRISE','YOLOv8'):

            from steren.networks.YOLOv8 import YOLOv8
            detector = YOLOv8(model_weights,
                                tile_size, 
                                overlap,
                                conf_thresh,
                                iou_thresh)

            hrimage = HRimage(img_path = img_path)
            hrimage.detect(detector) 
            hrimage.export_detections(output_file)





@app.command()
def slice(
    ctx: typer.Context,
    type: str= typer.Argument(help="Type of planetary data [Supported types: HiRISE]."),
    detector_type: str= typer.Argument(help="Type of detector [Supported types: YOLOv8]."),
    ):
    """
    Slice planetary data to become a regular image data set.
    """
    pass

if __name__ == "__main__":
    app()




from steren.network import Detector
import rasterio
import math
import os
import numpy as np
import cv2
import torch
from torchvision.ops import batched_nms
import rasterio
from rasterio.windows import Window
from rasterio.transform import Affine, xy
from shapely.geometry import box as shapely_box
import geopandas as gpd
from tqdm import tqdm

try:
    from ultralytics import YOLO
except ImportError:
    sys.exit("Ultralytics not found. Install with 'pip install ultralytics'.")





def make_tiles(img_w, img_h, tile_size, overlap):
    """Return list of (x1,y1,x2,y2) pixel coords covering image with overlap (pixels)."""
    step = tile_size - overlap
    nx = max(1, math.ceil((img_w - overlap) / step))
    ny = max(1, math.ceil((img_h - overlap) / step))

    tiles = []
    for iy in range(ny):
        for ix in range(nx):
            x1 = ix * step
            y1 = iy * step
            x2 = x1 + tile_size
            y2 = y1 + tile_size

            if x2 > img_w:
                x2 = img_w
                x1 = max(0, x2 - tile_size)
            if y2 > img_h:
                y2 = img_h
                y1 = max(0, y2 - tile_size)

            tiles.append((int(x1), int(y1), int(x2), int(y2)))
    # remove duplicates
    tiles = list(dict.fromkeys(tiles))
    return tiles


def read_tile_rgb(dataset, x1, y1, x2, y2, out_size=None):
    """
    Read window [x1,x2) x [y1,y2) from rasterio dataset and return HWC RGB uint8 array.
    out_size: (height, width) to rescale the read tile to (optional).
    Assumes raster has at least 3 bands (RGB). If >3 bands, uses first 3.
    """
    w = x2 - x1
    h = y2 - y1
    window = Window(x1, y1, w, h)
    # Read bands (1-indexed). Read minimum of 3 bands or all if fewer
    count = dataset.count
    bands_to_read = min(3, count)
    arr = dataset.read(range(1, bands_to_read + 1), window=window)  # shape (bands, h, w)

    # If bands < 3, replicate or pad
    if arr.shape[0] == 1:
        arr = np.repeat(arr, 3, axis=0)
    elif arr.shape[0] == 2:
        arr = np.vstack([arr, arr[0:1, :, :]])

    # convert to HWC
    tile = np.transpose(arr, (1, 2, 0))  # H, W, C

    # Raster values may be not in 0-255. Attempt rescale if dtype != uint8
    if tile.dtype != np.uint8:
        # scale using min/max per-tile (safe heuristic). If you want to use dataset-specific scaling, change here.
        tmin, tmax = tile.min(), tile.max()
        if tmax > tmin:
            tile = ((tile - tmin) / (tmax - tmin) * 255.0).astype(np.uint8)
        else:
            tile = np.clip(tile, 0, 255).astype(np.uint8)
    else:
        tile = tile.copy()

    # If out_size specified, resize tile to that size (width,height)
    if out_size is not None:
        tile = cv2.resize(tile, (out_size[1], out_size[0]), interpolation=cv2.INTER_LINEAR)

    # Ensure channel order is RGB (rasterio gives bands in dataset order)
    return tile




class YOLOv8(Detector):
    def __init__(self, 
                 weights :str, 
                 tile_size: int, 
                 overlap:int,
                 conf_thresh :float,
                 iou_thresh:float,
                 device:int = None
                 ):

        self.weights = weights
        self.tile_size = tile_size
        self.overlap = overlap
        self.iou_thresh = iou_thresh
        self.conf_thresh = conf_thresh

        self.model = YOLO(weights)
        if device:
            self.model.model.to(device)
    def set_device(self):
        pass 

    def load_weights(self):
        pass


    def inference(self,raster_path : str):
        
        if not os.path.exists(raster_path):
            raise FileNotFoundError(raster_path)

        with rasterio.open(raster_path) as ds:
            img_w = ds.width
            img_h = ds.height
            transform = ds.transform
            crs = ds.crs  # may be None

            tiles = make_tiles(img_w, img_h, self.tile_size, self.overlap)
            print(f"[info] image: {img_w}x{img_h}, tiles: {len(tiles)}, tile={self.tile_size}x{self.tile_size}, overlap={self.overlap}")
            all_boxes = []   # pixel-space boxes [x1,y1,x2,y2]
            all_scores = []
            all_classes = []

            for t_idx, (x1, y1, x2, y2) in enumerate(tqdm(tiles)):
                # Read tile into RGB np.uint8. Resize to model-friendly size if desired (we let model auto-resize)
                tile_rgb = read_tile_rgb(ds, x1, y1, x2, y2, out_size=None)  # HWC uint8
                if tile_rgb is None:
                    continue

                # Ultralytics model.predict accepts numpy HWC (RGB)
                results = self.model.predict(source=tile_rgb, 
                                             conf=self.conf_thresh, 
                                             imgsz=min(self.tile_size, 2048), 
                                             verbose=False)
                r = results[0]

                #if r.boxes is None or len(r.boxes) == 0:
                    #if debug:
                     #   print(f" tile {t_idx+1}/{len(tiles)}: 0 boxes")
                    #continue

                boxes = r.boxes.xyxy.cpu().numpy()  # coordinates in tile pixel space
                scores = r.boxes.conf.cpu().numpy()
                classes = r.boxes.cls.cpu().numpy().astype(int)

                # map tile coords -> full image pixel coords by adding offsets
                boxes[:, 0] += x1  # x1
                boxes[:, 1] += y1  # y1
                boxes[:, 2] += x1  # x2
                boxes[:, 3] += y1  # y2

                for b, s, c in zip(boxes, scores, classes):
                    all_boxes.append(b.tolist())
                    all_scores.append(float(s))
                    all_classes.append(int(c))

                #if debug:
               # print(f" tile {t_idx+1}/{len(tiles)}: {len(boxes)} boxes")

        # no detections
        if len(all_boxes) == 0:
            print("[info] no detections found")
            return

        # convert to tensors and clamp within image bounds
        boxes = torch.tensor(all_boxes, dtype=torch.float32)
        scores = torch.tensor(all_scores, dtype=torch.float32)
        labels = torch.tensor(all_classes, dtype=torch.int64)

        # clamp boxes to image
        boxes[:, 0].clamp_(0, img_w - 1)
        boxes[:, 1].clamp_(0, img_h - 1)
        boxes[:, 2].clamp_(0, img_w - 1)
        boxes[:, 3].clamp_(0, img_h - 1)

        # Perform class-aware NMS
        keep = batched_nms(boxes, scores, labels, self.iou_thresh)
        boxes_kept = boxes[keep].cpu().numpy()
        scores_kept = scores[keep].cpu().numpy()
        labels_kept = labels[keep].cpu().numpy()

        print(f"[info] {len(all_boxes)} raw boxes -> {len(keep)} after global NMS")

        
        return boxes_kept, scores_kept, labels_kept, crs, transform







from abc import ABC, abstractmethod
import numpy as np


class Network(ABC):
    @abstractmethod
    def load_weights(self):
        pass

    @abstractmethod
    def inference(self, image : np.ndarray):
        pass

    @abstractmethod
    def set_device(self, device : str):
        pass


class Detector(Network, ABC):
    pass 
    

class Segmenter(Network,ABC):
    pass
   

